
(define-type FAE
  [num (n number?)]
  [add (lhs FAE?)
       (rhs FAE?)]
  [sub (lhs FAE?)
       (rhs FAE?)]
  [id (name symbol?)]
  [fun (param symbol?)
       (body FAE?)]
  [app (fun-expr FAE?)
       (arg-expr FAE?)]
  [if0 (test FAE?)
       (then FAE?)
       (else FAE?)])

#|
(define-type CFAE
8  [cnum (n number?)]
9  [cadd (lhs CFAE?)
        (rhs CFAE?)]
10  [csub (lhs CFAE?)
        (rhs CFAE?)]
11  [cid (pos number?)]
12  [cfun (body CFAE?)]
13  [capp (fun-expr CFAE?)
       (arg-expr CFAE?)]
14  [cif0 (test CFAE?)
        (then CFAE?)
        (else CFAE?)])
|#

#|
(define-type CFAE-Value
15  [numV (n number?)]
16  [closureV (body CFAE?)
            (sc SubCache?)])
|#

(define-type CSubCache
  [mtCSub]
  [aCSub (name symbol?)
         (rest CSubCache?)])

#|
(define-type CFAE-Cont
0  [mtK]
1  [addSecondK (r CFAE?)
              (sc SubCache?)
              (k CFAE-Cont?)]
2  [doAddK (v1 CFAE-Value?)
          (k CFAE-Cont?)]
3  [subSecondK (r CFAE?)
              (sc SubCache?)
              (k CFAE-Cont?)]
4  [doSubK (v1 CFAE-Value?)
          (k CFAE-Cont?)]
5  [appArgK (arg-expr CFAE?)
           (sc SubCache?)
           (k CFAE-Cont?)]
6  [doAppK (fun-val CFAE-Value?)
          (k CFAE-Cont?)]
7  [doIfK (then-expr CFAE?)
         (else-expr CFAE?)
         (sc SubCache?)
         (k CFAE-Cont?)])
|#

#|
 17 cons for subcache
|#

;; A SubCache is a list of CFAE-Value
(define (SubCache? x) true)

;; ----------------------------------------
;; Allocation
;;  Instead of cons cells, use vectors
;;  This makes allocation a little more complex,
;;   but it will pay off in simplifying GC

(define memory (make-vector 1500))
(define ptr 0)

(define (incptr n)
  (begin
    (set! ptr (+ ptr n))
    (- ptr n)))

;; malloc1 : number number -> number
(define (malloc1 tag a)
  (begin
    (vector-set! memory ptr tag)
    (vector-set! memory (+ ptr 1) a)
    (incptr 2)))

;; malloc2 : number number number -> number
(define (malloc2 tag a b)
  (begin
    (vector-set! memory ptr tag)
    (vector-set! memory (+ ptr 1) a)
    (vector-set! memory (+ ptr 2) b)
    (incptr 3)))

;; malloc3 : number number number number -> number
(define (malloc3 tag a b c)
  (begin
    (vector-set! memory ptr tag)
    (vector-set! memory (+ ptr 1) a)
    (vector-set! memory (+ ptr 2) b)
    (vector-set! memory (+ ptr 3) c)
    (incptr 4)))

;; malloc4 : number number number number number -> number
(define (malloc4 tag a b c d)
  (begin
    (vector-set! memory ptr tag)
    (vector-set! memory (+ ptr 1) a)
    (vector-set! memory (+ ptr 2) b)
    (vector-set! memory (+ ptr 3) c)
    (vector-set! memory (+ ptr 4) d)
    (incptr 5)))

;; ref : number number -> number
(define (ref n d)
  (vector-ref memory (+ n d)))

;; ----------------------------------------

;; parse : S-expr -> FAE
(define (parse sexp)
  (cond
    [(number? sexp) (num sexp)]
    [(symbol? sexp) (id sexp)]
    [(pair? sexp)
     (case (car sexp)
       [(+) (add (parse (second sexp)) (parse (third sexp)))]
       [(-) (sub (parse (second sexp)) (parse (third sexp)))]
       [(fun) (fun (first (second sexp)) (parse (third sexp)))]
       [(if0) (if0 (parse (second sexp))
                   (parse (third sexp))
                   (parse (fourth sexp)))]
       [else (app (parse (first sexp)) (parse (second sexp)))])]))

(test (parse 3) (num 3))
(test (parse 'x) (id 'x))
(test (parse '{+ 1 2}) (add (num 1) (num 2)))
(test (parse '{- 1 2}) (sub (num 1) (num 2)))
(test (parse '{fun {x} x}) (fun 'x (id 'x)))
(test (parse '{1 2}) (app (num 1) (num 2)))
(test (parse '{if0 0 1 2}) (if0 (num 0) (num 1) (num 2)))

;; ----------------------------------------

;; compile : FAE CSubCache -> CFae
(define (compile a-fae sc)
  (type-case FAE a-fae
    [num (n) (malloc1 8 n)]
    [add (l r) (malloc2 9 (compile l sc) (compile r sc))]
    [sub (l r) (malloc2 10 (compile l sc) (compile r sc))]
    [id (name) (malloc1 11 (locate name sc))]
    [fun (param body-expr) (malloc1 12
                                    (compile body-expr (aCSub param sc)))]
    [app (fun-expr arg-expr)
         (malloc2 13 
                  (compile fun-expr sc)
                  (compile arg-expr sc))]
    [if0 (test-expr then-expr else-expr)
         (malloc3 14
                  (compile test-expr sc)
                  (compile then-expr sc)
                  (compile else-expr sc))]))

(define (locate name sc)
  (type-case CSubCache sc
    [mtCSub () (error 'locate "free variable")]
    [aCSub (sub-name rest-sc)
           (if (symbol=? sub-name name)
               0
               (+ 1 (locate name rest-sc)))]))

;; ----------------------------------------

(define fae-reg 0)
(define sc-reg 0)

;; interp : FAE SubCache FAE-Cont -> alpha
(define (interp)
  (case (ref fae-reg 0)
    [(8) ; num
     (begin
       (set! v-reg (malloc1 15 (ref fae-reg 1)))
       (continue))]
    [(9) ; add
     (begin
       (set! k-reg (malloc3 1
                            (ref fae-reg 2)
                            sc-reg 
                            k-reg))
       (set! fae-reg (ref fae-reg 1))
       (interp))]
    [(10) ; sub
     (begin
       (set! k-reg (malloc3 3
                            (ref fae-reg 2)
                            sc-reg k-reg))
       (set! fae-reg (ref fae-reg 1))
       (interp))]
    [(11) ; id
     (begin
       (set! sc2-reg sc-reg)
       (set! v-reg (ref fae-reg 1))
       (sc-ref))]
    [(12) ; fun
     (begin
       (set! v-reg (malloc2 16 (ref fae-reg 1) sc-reg))
       (continue))]
    [(13) ; app
     (begin
       (set! k-reg (malloc3 5
                            (ref fae-reg 2)
                            sc-reg k-reg))
       (set! fae-reg (ref fae-reg 1))
       (interp))]
    [(14) ; if0
     (begin
       (set! k-reg (malloc4 7
                            (ref fae-reg 2)
                            (ref fae-reg 3)
                            sc-reg k-reg))
       (set! fae-reg (ref fae-reg 1))
       (interp))]))

(define k-reg 0)
(define v-reg 0)

;; continue : -> void
(define (continue)
  (case (ref k-reg 0)
    [(0) ; mtk
     v-reg]
    [(1) ; addSecondK
     (begin
       (set! fae-reg (ref k-reg 1))
       (set! sc-reg (ref k-reg 2))
       (set! k-reg (malloc2 2 v-reg (ref k-reg 3)))
       (interp))]
    [(2) ; doAddK
     (begin
       (set! v-reg (num+ (ref k-reg 1) v-reg))
       (set! k-reg (ref k-reg 2))
       (continue))]
    [(3) ; subSecondK
     (begin
       (set! fae-reg (ref k-reg 1))
       (set! sc-reg (ref k-reg 2))
       (set! k-reg (malloc2 4 v-reg (ref k-reg 3)))
       (interp))]
    [(4) ; doSubK
     (begin
       (set! v-reg (num- (ref k-reg 1) v-reg))
       (set! k-reg (ref k-reg 2))
       (continue))]
    [(5) ; appArgK
     (begin
       (set! fae-reg (ref k-reg 1))
       (set! sc-reg (ref k-reg 2))
       (set! k-reg (malloc2 6 v-reg (ref k-reg 3)))
       (interp))]
    [(6) ; doAppK
     (begin
       (set! fae-reg (ref (ref k-reg 1) 1))
       (set! sc-reg (malloc2 17
                             v-reg
                             (ref (ref k-reg 1) 2)))
       (set! k-reg (ref k-reg 2))
       (interp))]
    [(7) ; doIfK
     (begin
       (if (numzero? v-reg)
           (set! fae-reg (ref k-reg 1))
           (set! fae-reg (ref k-reg 2)))
       (set! sc-reg (ref k-reg 3))
       (set! k-reg (ref k-reg 4))
       (interp))]))

;; num-op : (number number -> number) -> (FAE-Value FAE-Value -> FAE-Value)
(define (num-op op op-name)
  (lambda (x y)
    (malloc1 15 (op (ref x 1) (ref y 1)))))

(define num+ (num-op + '+))
(define num- (num-op - '-))

;; numzero? : FAE-Value -> boolean
(define (numzero? x)
  (zero? (ref x 1)))

(define sc2-reg 0)

(define (sc-ref)
  (if (zero? v-reg)
      (begin
        (set! v-reg (ref sc2-reg 1))
        (continue))
      (begin
        (set! sc2-reg (ref sc2-reg 2))
        (set! v-reg (- v-reg 1))
        (sc-ref))))

;; ----------------------------------------

(define init-k (malloc1 0 0))
(define (interpx a-fae sc k)
  (begin
    (set! fae-reg a-fae)
    (set! sc-reg sc)
    (set! k-reg k)
    (interp)))
(define (numV x) (malloc1 15 x))
(define empty-sc 0)

(define (vtest a b)
  (test (ref a 1) (ref b 1)))

(vtest (interpx (compile (num 10) (mtCSub))
                empty-sc
                init-k)
       (numV 10))
(vtest (interpx (compile (add (num 10) (num 7)) (mtCSub))
                empty-sc
                init-k)
       (numV 17))
(vtest (interpx (compile (sub (num 10) (num 7)) (mtCSub))
                empty-sc
                init-k)
       (numV 3))
(vtest (interpx (compile
                 (app (fun 'x (add (id 'x) (num 12)))
                      (add (num 1) (num 17)))
                 (mtCSub))
                empty-sc
                init-k)
       (numV 30))
(vtest (interpx (compile (id 'x)
                         (aCSub 'x (mtCSub)))
                (malloc2 17 (numV 10) empty-sc)
                init-k)
       (numV 10))

(vtest (interpx (compile (app (fun 'x (add (id 'x) (num 12)))
                              (add (num 1) (num 17)))
                         (mtCSub))
                empty-sc
                init-k)
       (numV 30))

(vtest (interpx (compile (app (fun 'x
                                   (app (fun 'f
                                             (add (app (id 'f) (num 1))
                                                  (app (fun 'x
                                                            (app (id 'f) (num 2)))
                                                       (num 3))))
                                        (fun 'y (add (id 'x) (id 'y)))))
                              (num 0))
                         (mtCSub))
                empty-sc
                init-k)
       (numV 3))

(vtest (interpx (compile (if0 (num 0)
                              (num 1)
                              (num 2))
                         (mtCSub))
                empty-sc
                init-k)
       (numV 1))
(vtest (interpx (compile (if0 (num 1)
                              (num 0)
                              (num 2))
                         (mtCSub))
                empty-sc
                init-k)
       (numV 2))

(vtest (interpx (compile
                 (parse 
                  '{{fun {mkrec}
                         {{fun {fib}
                               ;; Call fib on 4:
                               {fib 4}}
                          ;; Create recursive fib:
                          {mkrec
                           {fun {fib}
                                ;; Fib:
                                {fun {n}
                                     {if0 n
                                          1
                                          {if0 {- n 1}
                                               1
                                               {+ {fib {- n 1}}
                                                  {fib {- n 2}}}}}}}}}}
                    ;; mkrec:
                    {fun {body-proc}
                         {{fun {fX}
                               {fX fX}}
                          {fun {fX}
                               {body-proc {fun {x} {{fX fX} x}}}}}}})
                 (mtCSub))
                empty
                init-k)
       (numV 5))
